"""
This module defines base controls, which are interactive blocks.
"""
import threading
from typing import Any, Dict

from .block import Block
from .common import Commons


class BorderStyle:
    """
    Defines a border character set.

    Attributes
    ----------
    chars : list(str)
        A 6-element list which contains the border characters for the top/bottom, left/right sides, followed by the
        top left, top right, bottom left and bottom right corners.
    """

    def __init__(self, chars=None):
        """
        Initializes a BorderStyle object.

        Parameters
        ----------
        chars : list(str)
            A 6-element list which contains the border characters for the top/bottom, left/right sides, followed by the
            top left, top right, bottom left and bottom right corners.
        """
        if chars is None:
            chars = ["-", "|", "-", "-", "-", "-"]
        self.chars = chars[:]

    @property
    def horizontal(self):
        """
        Read-only property for the horizontal border character.

        Returns
        -------
        str
            The horizontal border character.
        """
        return self.chars[0]

    @property
    def vertical(self):
        """
        Read-only property for the vertical border character.

        Returns
        -------
        str
            The vertical border character.
        """
        return self.chars[1]

    @property
    def topleft(self):
        """
        Read-only property for the top left corner border character.

        Returns
        -------
        str
            The top left corner border character.
        """
        return self.chars[2]

    @property
    def topright(self):
        """
        Read-only property for the top right corner border character.

        Returns
        -------
        str
            The top right corner border character.
        """
        return self.chars[3]

    @property
    def bottomleft(self):
        """
        Read-only property for the bottom left corner border character.

        Returns
        -------
        str
            The bottom left corner border character.
        """
        return self.chars[4]

    @property
    def bottomright(self):
        """
        Read-only property for the bottom right corner border character.

        Returns
        -------
        str
            The bottom right corner border character.
        """
        return self.chars[5]


BorderStyleContinuous = BorderStyle(["─", "│", "┌", "┐", "└", "┘"])
"""
Border style preset with a set of characters that form a continuous border.
"""


class Border:
    """
    Defines the border of a control.

    Attributes
    ----------
    style : awsc.termui.control.BorderStyle
        The border style character set.
    color : awsc.termui.color.Color
        The color of the border.
    title : strs
        The main title to display in the title bar of the border.
    title_color : awsc.termui.color.Color
        The color of the main title text.
    title_info : str
        The additional information to display in the title bar of the border.
    title_info_color : awsc.termui.color.Color
        The color of the additional title info text.
    """

    def __init__(
        self,
        style=None,
        color=None,
        title=None,
        title_color=None,
        title_info=None,
        title_info_color=None,
    ):
        """
        Initializes a Border object.

        Parameters
        ----------
        style : awsc.termui.control.BorderStyle
            The border style character set.
        color : awsc.termui.color.Color
            The color of the border.
        title : str, optional
            The main title to display in the title bar of the border.
        title_color : awsc.termui.color.Color, optional
            The color of the main title text.
        title_info : str, optional
            The additional information to display in the title bar of the border.
        title_info_color : awsc.termui.color.Color, optional
            The color of the additional title info text.
        """
        self.style = style
        self.color = color
        self.title = title
        self.title_color = title_color
        self.title_info = title_info
        self.title_info_color = title_info_color

    def paint(self, block):
        """
        Paint hook for a border. Expected to output the border within the bounds of the parameter block.

        Parameters
        ----------
        block : awsc.termui.block.Block
            The block which this border is painted on.
        """
        if self.style is None or self.color is None:
            return
        corners = block.corners
        dimensions = block.dimensions()
        for i in range(corners[1][0], corners[1][1] + 1):
            if i == corners[1][0]:
                if self.title is None or self.title_color is None:
                    outer = (
                        self.style.topleft
                        + self.style.horizontal * (dimensions[0] - 1)
                        + self.style.topright
                    )
                    Commons.UIInstance.print(
                        outer, xy=(corners[0][0], i), color=self.color
                    )
                else:
                    total_title = self.title
                    if self.title_info is not None:
                        total_title = f"{self.title} ({self.title_info})"
                    cpos = int((dimensions[0] + 1) / 2)
                    spos = cpos - int(len(total_title) / 2) - 1
                    slen = len(total_title) + 2
                    outer = self.style.topleft + self.style.horizontal * spos
                    length = len(outer)
                    Commons.UIInstance.print(
                        outer, xy=(corners[0][0], i), color=self.color
                    )
                    outer = " " + self.title + " "
                    Commons.UIInstance.print(
                        outer,
                        xy=(corners[0][0] + length, i),
                        color=self.title_color,
                        bold=True,
                    )
                    length += len(outer)
                    if self.title_info is not None:
                        Commons.UIInstance.print(
                            "(",
                            xy=(corners[0][0] + length, i),
                            color=self.title_color,
                            bold=True,
                        )
                        length += 1
                        Commons.UIInstance.print(
                            self.title_info,
                            xy=(corners[0][0] + length, i),
                            color=self.title_color
                            if self.title_info_color is None
                            else self.title_info_color,
                            bold=True,
                        )
                        length += len(self.title_info)
                        Commons.UIInstance.print(
                            ") ",
                            xy=(corners[0][0] + length, i),
                            color=self.title_color,
                            bold=True,
                        )
                        length += 2
                    outer = (
                        self.style.horizontal * (dimensions[0] - 1 - spos - slen)
                        + self.style.topright
                    )
                    Commons.UIInstance.print(
                        outer, xy=(corners[0][0] + length, i), color=self.color
                    )
            elif i == corners[1][1]:
                outer = (
                    self.style.bottomleft
                    + self.style.horizontal * (dimensions[0] - 1)
                    + self.style.bottomright
                )
                Commons.UIInstance.print(outer, xy=(corners[0][0], i), color=self.color)
            else:
                outer = (
                    self.style.vertical
                    + " " * (dimensions[0] - 1)
                    + self.style.vertical
                )
                Commons.UIInstance.print(outer, xy=(corners[0][0], i), color=self.color)


BorderNone = Border()
"""
A preset which defines a border which paints nothing.
"""


class Control(Block):
    """
    The base object for all controllable blocks.

    Attributes
    ----------
    _border : awsc.termui.control.Border
        The border of the control.
    thread_share : dict
        A cross-thread storage for update threads to share data with the main execution thread.
    mutex : threading.Lock
        A mutex used to lock the cross-thread storage.
    """

    def __init__(self, *args, border=None, **kwargs):
        """
        Initializes a Control object.

        Parameters
        ----------
        border : awsc.termui.control.Border
            The border of the control.
        """
        super().__init__(*args, **kwargs)
        self.border = border
        self.thread_share = {}
        self.mutex = threading.Lock()

    @property
    def border(self):
        """
        Returns the border of the control.

        Returns
        -------
        awsc.termui.control.Border
            The border of the control.
        """
        return self._border

    @border.setter
    def border(self, value):
        """
        Sets the border of the control and forces a UI repaint.

        Parameters
        -------
        value : awsc.termui.control.Border
            The border of the control.
        """
        self._border = value
        Commons.UIInstance.dirty = True

    @property
    def inner(self):
        corners = self.corners
        x0 = corners[0][0] + (0 if self.border is None else 1)
        x1 = corners[0][1] - (0 if self.border is None else 1)
        y0 = corners[1][0] + (0 if self.border is None else 1)
        y1 = corners[1][1] - (0 if self.border is None else 1)
        return ((x0, x1), (y0, y1))

    @property
    def w_in(self):
        return self.width if self.border is None else self.width - 2

    def paint(self):
        if self.border is not None:
            self.border.paint(self)
        super().paint()


class HotkeyControl(Control):
    """
    Base class for all controls which implement a hotkey system.

    Attributes
    ----------
    hotkeys : dict
        A mapping of hotkeys to hotkey callbacks.
    tooltips : dict
        A mapping of hotkeys to short descriptions of their actions.
    hotkey_is_validated : dict
        A mapping of hotkey to bools. If a hotkey is validated, the validate_hotkey method will be called with the hotkey and it must
        return True for the hotkey to pass through to the input method.
    """

    class Autohotkey:
        """
        Decorator that automatically assigns a hotkey to a function upon instantiation.

        This is fuelled by the worst python can offer, I wish you the best of luck in understanding its source.

        Usage: @HotkeyControl.Autohotkey(key, tooltip=None, is_validated=False)

        HotkeyControl.Autohotkey may only be used for methods in classes which extend HotkeyControl.

        Attributes
        ----------
        key : str
            The hotkey to react to. See HotkeyControl.add_hotkey for more details.
        tooltip : str, optional
            Tooltip to associate with the hotkey. See HotkeyControl.add_hotkey for more details.
        is_validated : bool, default=False
            Is the hotkey validated? See HotkeyControl.add_hotkey for more details.
        condition : bool, default=False
            Run the condition callback of the HotkeyControl before adding.
        """

        _autohotkeys: Dict[str, Dict[str, Dict[str, Any]]] = {}

        def __init__(self, key, tooltip=None, is_validated=False, condition=False):
            self.key = key
            self.tooltip = tooltip
            self.is_validated = is_validated

        def __call__(self, function):
            definer = function.__qualname__.split(".")[0]
            if definer not in self.__class__._autohotkeys:
                self.__class__._autohotkeys[definer] = {}
            self.__class__._autohotkeys[definer][self.key] = {
                "callback": function,
                "tooltip": self.tooltip,
                "is_validated": self.is_validated,
            }
            return function

        @classmethod
        def auto_hotkey(cls, self):
            """
            Performs automatically hotkeying up a HotkeyControl subclass.

            Inserts all inherited hotkeys automatically. Insertion is done in reverse of MRO - therefore, higher classes on the inheritance tree,
            such as HotkeyControl, insert their hotkeys first. This allows subclasses to overwrite these hotkeys.

            Parameters
            ----------
            self : awsc.termui.control.HotkeyControl
                The class being initialized.
            """
            import inspect

            for classobj in reversed(inspect.getmro(self.__class__)):
                classname = classobj.__name__
                if classname in cls._autohotkeys:
                    for hotkey, definition in cls._autohotkeys[classname].items():
                        # pylint: disable=unnecessary-dunder-call # Offering a bounty of a single point of reddit karma if anyone can tell me why.
                        self.add_hotkey(
                            hotkey,
                            definition["callback"].__get__(self, self.__class__),
                            definition["tooltip"],
                            definition["is_validated"],
                        )

    def __init__(self, *args, **kwargs):
        """
        Initializes a HotkeyControl object.
        """
        super().__init__(*args, **kwargs)
        self.hotkeys = {}
        self.tooltips = {}
        self.hotkey_is_validated = {}
        HotkeyControl.Autohotkey.auto_hotkey(self)

    def autohotkey_condition(self, hotkey):
        """
        Autohotkey condition callback. Called with the key name if the key is conditional.

        Parameters
        ----------
        hotkey : str
            The key definition.

        Returns
        -------
        bool
            True if condition is met.
        """
        return True

    def add_hotkey(self, hotkey, action, tooltip=None, is_validated=False):
        """
        Adds a new hotkey to the control.

        Parameters
        ----------
        hotkey : str
            The hotkey for the action being added. Hotkeys can be a number of things:
            * A character, such as 'a' or '.', directly equivalent to pressing the associated key. Not case sensitive.
            * A control code from awsc.termui.ui.ControlCodes. Control codes are the equivalent of holding the CTRL key while pressing the hotkey,
              so ControlCodes.A is CTRL+A. Note that due to how the terminal handles control codes (^A being 0, ^B being 1, and so on), certain
              control codes such as CTRL+J are unavailable, as their codes are equivalent to the codes of keys that type special characters, such
              as ENTER or TAB.
            * A key name. Special keys may be referred to by name, such as KEY_ESCAPE or KEY_ENTER. A list of key names is available in the
              documentation of the blessed library (https://blessed.readthedocs.io/en/latest/keyboard.html).
        action : callable(object)
            The callback to call if the hotkey is pressed and optionally passes validation. The callback is passed a single parameter, which is
            this object and does not need to return a value.
        tooltip : str, optional
            An optional tooltip associated with the hotkey. This may be used by other controls to create hotkey displays.
        is_validated: bool, default=False
            If True, hotkey must pass validation before being handled. Validation is implemented by the control in the validate_hotkey method.
        """
        self.hotkeys[hotkey] = action
        if tooltip is not None:
            self.tooltips[hotkey] = tooltip
        self.hotkey_is_validated[hotkey] = is_validated
        Commons.UIInstance.dirty = True

    def validate_hotkey(self, key):
        """
        Validates whether a hotkey can be used based on the state of the control.

        Parameters
        ----------
        key : str
            The hotkey being used, which is either a lowercase character, a control code or a key name based on which key was pressed.

        Returns
        -------
        bool
            Returns True if the hotkey is usable, False to prevent using it.
        """
        return True

    def input(self, key):
        inkey = str(key)
        if key.is_sequence:
            inkey = key.name
        else:
            inkey = inkey.lower()
        if inkey in self.hotkeys:
            if not self.hotkey_is_validated[inkey] or self.validate_hotkey(inkey):
                self.hotkeys[inkey](self)
                Commons.UIInstance.dirty = True
            return True
        return False



"""
Module for keypair resources.
"""
from pathlib import Path

from .base_control import Describer, ResourceLister
from .common import Common, SessionAwareDialog
from .ssh import SSHList
from .termui.control import Border
from .termui.dialog import DialogFieldLabel, DialogFieldText


class KeyPairDescriber(Describer):
    """
    Describer control for keypairs.
    """

    prefix = "keypair_browser"
    title = "EC2 Keypair"

    resource_type = "keypair"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "Keypair"
    describe_method = "describe_key_pairs"
    describe_kwarg_name = "KeyPairIds"
    describe_kwarg_is_list = True
    object_path = ".KeyPairs[0]"
    default_entry_key = "id"


def _keypair_determine_association(entry):
    return Common.Session.get_keypair_association(entry["KeyPairId"])


class KeyPairResourceLister(ResourceLister):
    """
    Lister control for keypairs.
    """

    prefix = "keypair_list"
    title = "EC2 Keypairs"
    command_palette = ["key", "keypair"]

    resource_type = "keypair"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "Keypair"
    list_method = "describe_key_pairs"
    item_path = ".KeyPairs"
    columns = {
        "id": {
            "path": ".KeyPairId",
            "size": 16,
            "weight": 0,
        },
        "name": {"path": ".KeyName", "size": 32, "weight": 1, "sort_weight": 0},
        "fingerprint": {
            "path": ".KeyFingerprint",
            "size": 48,
            "weight": 2,
        },
        "association": {
            "path": _keypair_determine_association,
            "size": 16,
            "weight": 3,
        },
    }
    primary_key = "id"
    describe_command = KeyPairDescriber.opener

    @ResourceLister.Autohotkey("a", "Associate with SSH key", True)
    def set_keypair_association(self, _):
        """
        Hotkey callback for setting keypair association.
        """
        Common.Session.push_frame(SSHList.selector(self.ssh_selector_callback))
        Common.Session.ui.dirty = True

    def ssh_selector_callback(self, val):
        """
        Selector callback for keypair association.
        """
        Common.Session.set_keypair_association(self.selection["id"], val)
        self.refresh_data()

    @ResourceLister.Autohotkey("n", "New keypair")
    def create_keypair(self, _):
        """
        Hotkey callback for creating a new keypair.
        """
        KeyPairCreateDialog.opener(self, caller=self)

    @ResourceLister.Autohotkey("x", "Remove association", True)
    def remove_keypair_association(self, _):
        """
        Hotkey callback for deleting keypair association.
        """
        Common.Session.set_keypair_association(self.selection["id"], "")
        self.refresh_data()


class KeyPairCreateDialog(SessionAwareDialog):
    """
    Dialog for creating a keypair.

    Attributes
    ----------
    name_field : awsc.termui.dialog.DialogFieldText
        The textfield for entering a name.
    error_label : awsc.termui.dialog.DialogFieldLabel
        A label where error feedback may be dispalyed.
    dotssh : pathlib.Path
        The path to ~/.ssh
    save_as_field : awsc.termui.dialog.DialogFieldText
        The name of the output file for the downloaded SSH key.
    """

    def __init__(self, *args, caller=None, **kwargs):
        kwargs["border"] = Border(
            Common.border("default"),
            Common.color("modal_dialog_border"),
            "Create new keypair",
            Common.color("modal_dialog_border_title"),
        )
        super().__init__(*args, caller=caller, **kwargs)
        self.set_title_label("Enter keypair details")
        self.name_field = DialogFieldText(
            "Name:",
            **Common.textfield_colors("ec2_keypair"),
        )
        self.add_field(self.name_field)
        self.dotssh = Path.home() / ".ssh"
        self.add_field(
            DialogFieldLabel(f"Key file will be saved in {str(self.dotssh.resolve())}")
        )
        self.save_as_field = DialogFieldText(
            "Filename:",
            **Common.textfield_colors("ec2_keypair"),
        )
        self.add_field(self.save_as_field)
        self.caller = caller

    def accept_and_close(self):
        if self.name_field.text == "":
            self.error_label.text = "Name cannot be blank."
            return
        if self.save_as_field.text == "":
            self.error_label.text = "Filename cannot be blank."
            return
        if "/" in self.save_as_field.text or "\\" in self.save_as_field.text:
            self.error_label.text = "Filename cannot contain path separators."
            return

        resps = Common.generic_api_call(
            "ec2",
            "create_key_pair",
            {"KeyName": self.name_field.text},
            "Create Keypair",
            "EC2",
            subcategory="Key Pair",
            success_template="Creating keypair {0}",
            resource=self.name_field.text,
        )
        if resps["Success"]:
            resp = resps["Response"]
            data = resp["KeyMaterial"]
            with (self.dotssh / self.save_as_field.text).open("w") as file:
                file.write(data)
            Common.Session.set_keypair_association(
                resp["KeyPairId"], self.save_as_field.text
            )

        super().accept_and_close()

    def close(self):
        if self.caller is not None:
            self.caller.refresh_data()
        super().close()



"""
Module for the encrypted key store object.
"""
import datetime
import getpass
import hashlib
import os
import sys

import yaml
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import AES
from cryptography.hazmat.primitives.ciphers.modes import CBC
from cryptography.hazmat.primitives.padding import PKCS7


class Keystore:
    """
    Encrypted key storage for AWS credentials.

    Attributes
    ----------
    keylist_file : pathlib.Path
        Encrypted binary which stores the AWS credentials.
    nonce_file : pathlib.Path
        Nonce file for the encryption.
    nonce : bytes
        The nonce value.
    keylist_cipher : cryptography.hazmat.primitives.cipher.Cipher
        The cryptography cipher used to decrypt and encrypt the key storage.
    keys : dict
        A list of decrypted key pairs, keyed by account name.
    """

    def __init__(self, config):
        """
        Initializes a Keystore object.

        Parameters
        ----------
        config : awsc.config.config.Configuration
            The parent configuration object instance.
        """
        self.keylist_file = config.path / "keys"
        self.nonce_file = config.path / "nonce"
        self.keylist_cipher = None
        self.nonce = os.urandom(16)
        self.keys = {}

    def do_unlock(self, password):
        """
        Unlock keystore with a password.
        """
        sha = hashlib.sha256(password.encode("ascii"))
        if not self.nonce_file.exists():
            with self.nonce_file.open("wb") as file:
                file.write(self.nonce)
        else:
            with self.nonce_file.open("rb") as file:
                self.nonce = file.read()
        self.keylist_cipher = Cipher(AES(sha.digest()), CBC(self.nonce))

        if self.keylist_file.exists():
            self.parse_keylist()

    def unlock(self, silent=False):
        """
        Prompts the user to enter the password to access the key storage, and parses the keylist file.
        """
        password = getpass.getpass(
            "Enter encryption phrase for key database: ",
            stream=sys.stderr if silent else None,
        )
        self.do_unlock(password)

    def parse_keylist(self):
        """
        Attempts to parse the keylist file.

        Should be called through unlock(), as unlock() generates the cipher for this method.
        """
        try:
            with self.keylist_file.open("rb") as file:
                data = file.read()
            dec = self.keylist_cipher.decryptor()
            yaml_decoded = dec.update(data) + dec.finalize()
            unpadder = PKCS7(256).unpadder()
            yaml_decoded = unpadder.update(yaml_decoded) + unpadder.finalize()
            self.keys = yaml.safe_load(yaml_decoded.decode("UTF-8"))
            changed = False
            for key in self.keys:
                if "temp" in self.keys[key]:
                    changed = True
                    self.keys[key].pop("temp")
            if changed:
                self.write_keylist()
        except ValueError:
            print("Incorrect password.")
            sys.exit(1)

    def write_keylist(self):
        """
        Writes the keylist to the keylist_file. Cipher must be loaded through unlock() before writing.
        """
        enc = self.keylist_cipher.encryptor()
        yaml_encoded = yaml.dump(self.keys).encode("UTF-8")
        padder = PKCS7(256).padder()
        padded = padder.update(yaml_encoded) + padder.finalize()
        data = enc.update(padded) + enc.finalize()
        with self.keylist_file.open("wb") as file:
            file.write(data)

    def __getitem__(self, item):
        """
        Returns the named keypair. Resolves references unless temporary credentials are available.

        Parameters
        ----------
        item : str
            Name of the keypair to fetch.

        Returns
        -------
        dict
            A dict containing the access and secret keys.
        """
        while key := self.keys[item]:
            if "temp" in key:
                if (
                    datetime.datetime.fromtimestamp(key["temp"]["expiry"])
                    > datetime.datetime.now()
                ):
                    return key["temp"]
            if "ref" in key:
                item = key["ref"]
                continue
            return key

    def force_resolve(self, item):
        """
        Returns the named keypair. Resolves references and ignored temporary credentials.

        Parameters
        ----------
        item : str
            Name of the keypair to fetch.

        Returns
        -------
        dict
            A dict containing the access and secret keys.
        """
        while key := self.keys[item]:
            if "ref" in key:
                item = key["ref"]
                continue
            if "temp" in key:
                if (
                    datetime.datetime.fromtimestamp(key["temp"]["expiry"])
                    > datetime.datetime.now()
                ):
                    return key["temp"]
            return key

    def get_permanent_credentials(self, item):
        """
        Returns the permanent credentials associated with a context.

        Parameters
        ----------
        item : str
            Name of the key..

        Returns
        -------
        dict
            Gets the permanent credentials associated with a context.
        """
        while key := self.keys[item]:
            if "ref" in key:
                item = key["ref"]
                continue
            return key

    def get_ref_name(self, item):
        """
        Returns the ref of the named keypair, as this property is not accessible through __getitem__.

        Parameters
        ----------
        item: str
            Name of the key with a reference.

        Returns
        -------
        str | None
            Name of the key being referenced, or None if none.
        """
        key = self.keys[item]
        return key["ref"] if "ref" in key else None

    def get_all_ref_targets(self):
        """
        Enumerates all ref targets into a list.

        Returns
        -------
        set(str)
            A list of keys which are targets of refs.
        """
        targets = set()
        for key in self.keys:
            if "ref" in key:
                targets.add(key["ref"])
        return targets

    def __contains__(self, item):
        return item in self.keys

    def set_key(self, name, access, secret):
        """
        Upserts a key into the key storage and writes to disk.

        Parameters
        ----------
        name : str
            The name of the keypair.
        access : str
            The access key of the keypair.
        secret : str
            The secret key of the keypair.
        """
        self.keys[name] = {"access": access, "secret": secret}
        self.write_keylist()

    def set_ref(self, name, ref):
        """
        Upserts a reference key into the key storage and writes to disk.

        Parameters
        ----------
        name : str
            The name of the keypair.
        ref : str
            The target keypair being referenced.
        """
        self.keys[name] = {"ref": ref}
        self.write_keylist()

    def set_temp(self, name, data):
        """
        Sets temporary credentials for a reference key.

        Parameters
        ----------
        name : str
            The name of the keypair.
        data : dict
            The temporary data of the keypair.
        """
        self.keys[name]["temp"] = data
        self.write_keylist()

    def delete_key(self, name):
        """
        Deletes a key from the key storage and writes to disk.

        Parameters
        ----------
        name : str
            The name of the key to delete.
        """
        del self.keys[name]
        self.write_keylist()



"""
Controls related to SSH keys.
"""
import re
from pathlib import Path

from .base_control import OpenableListControl
from .common import Common
from .termui.alignment import CenterAnchor, Dimension
from .termui.control import Border
from .termui.dialog import DialogControl, DialogFieldText
from .termui.list_control import ListEntry


# TODO: Implement usage frequency.
class SSHList(OpenableListControl):
    """
    Lister control for available SSH keys.
    """

    prefix = "ssh"
    title = "SSH Keys"

    def __init__(self, *args, selector_cb=None, **kwargs):
        super().__init__(
            *args,
            color=Common.color("ssh_key_list_generic", "generic"),
            selection_color=Common.color("ssh_key_list_selection", "selection"),
            title_color=Common.color("ssh_key_list_heading", "column_title"),
            selector_cb=selector_cb,
            **kwargs,
        )

        if selector_cb is None:
            self.add_hotkey("KEY_ENTER", self.select_ssh_key, "Select ssh key")
        self.add_column("usage frequency", 12)
        self.add_column("has public key", 12)
        self.add_column("default username", 20)
        self.add_column("default", 8)
        self.reload()

    def reload(self, move=True):
        """
        Reloads the list of available SSH keys.

        Parameters
        ----------
        move : bool
            If True, moves the selection cursor to the default SSH key.
        """
        self.entries = []
        idx = 0
        defa = 0
        for ssh_key in sorted(self.list_ssh_keys()):
            if ssh_key == Common.Configuration["default_ssh_key"]:
                defa = idx
                is_default = "✓"
            else:
                is_default = " "
            pubkey = Path.home() / ".ssh" / f"{ssh_key}.pub"
            has_pubkey = "✓" if pubkey.exists() else " "
            default_username = (
                Common.Configuration["default_ssh_usernames"][ssh_key]
                if ssh_key in Common.Configuration["default_ssh_usernames"]
                else ""
            )
            self.add_entry(
                ListEntry(
                    ssh_key,
                    **{
                        "usage frequency": 0,
                        "has public key": has_pubkey,
                        "default username": default_username,
                        "default": is_default,
                    },
                )
            )
            idx += 1
        if move:
            self.selected = defa

    def list_ssh_keys(self):
        """
        Fetches a list of available SSH keys from ~/.ssh. Only files beginning with `id_` or with the .pem extension are used.

        Returns
        -------
        list(str)
            A list of filenames in ~/.ssh matching the pattern.
        """
        ssh_dir = Path.home() / ".ssh"
        ret = []
        for child in ssh_dir.iterdir():
            filename = child.name
            if re.match(r"^((id_[^.]+?)|(.*?\.pem))$", filename) is not None:
                ret.append(filename)
        return ret

    @OpenableListControl.Autohotkey("d", "Set as default", True)
    def set_default_ssh_key(self, _):
        """
        Hotkey callback for setting the default SSH key.
        """
        if self.selection is not None:
            Common.Configuration["default_ssh_key"] = self.selection.name
            Common.Configuration.write_config()
            for entry in self.entries:
                if entry.name != Common.Configuration["default_ssh_key"]:
                    entry["default"] = " "
                else:
                    entry["default"] = "✓"

    def select_ssh_key(self, _):
        """
        Hotkey callback for setting the active SSH key.
        """
        if self.selection is not None:
            Common.Session.ssh_key = self.selection.name

    @OpenableListControl.Autohotkey("e", "Set default username", True)
    def set_default_ssh_username(self, _):
        """
        Hotkey callback for associating a default username with an SSH key.
        """
        SSHDefaultUsernameDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            key_name=self.selection["name"],
            caller=self,
            weight=-500,
        )


# TODO: Replace with SingleNameDialog?
class SSHDefaultUsernameDialog(DialogControl):
    """
    Dialog for entering the default username for an SSH key.

    Attributes
    ----------
    username_textfield : awsc.termui.dialog.DialogFieldText
        Textfield for entering the default username.
    caller : awsc.termui.control.Control
        The parent control controlling this dialog.
    key_name : str
        The SSH key to associate the username with.
    """

    def __init__(self, *args, key_name="", caller=None, **kwargs):
        self.key_name = key_name
        kwargs["ok_action"] = self.accept_and_close
        kwargs["cancel_action"] = self.close
        kwargs["border"] = Border(
            Common.border("ssh", "default"),
            Common.color("ssh_modal_dialog_border", "modal_dialog_border"),
            "Default SSH username",
            Common.color("ssh_modal_dialog_border_title", "modal_dialog_border_title"),
            self.key_name,
            Common.color(
                "ssh_modal_dialog_border_title_info", "modal_dialog_border_title_info"
            ),
        )
        super().__init__(*args, **kwargs)
        def_text = (
            Common.Configuration["default_ssh_usernames"][self.key_name]
            if self.key_name in Common.Configuration["default_ssh_usernames"]
            else ""
        )
        self.username_textfield = DialogFieldText(
            "SSH username",
            text=def_text,
            **Common.textfield_colors("ec2_ssh"),
        )
        self.add_field(self.username_textfield)
        self.highlighted = 1 if def_text != "" else 0
        self.caller = caller

    def input(self, key):
        if key.is_sequence and key.name == "KEY_ESCAPE":
            self.close()
            return True
        return super().input(key)

    def accept_and_close(self):
        """
        Affirmative action callback. Sets the ssh key default username.
        """
        Common.Configuration["default_ssh_usernames"][
            self.key_name
        ] = self.username_textfield.text
        Common.Configuration.write_config()
        Common.Session.ssh_key = Common.Session.ssh_key  # don't ask
        self.caller.reload(move=False)
        self.close()

    def close(self):
        """
        Negative action callback. Closes the dialog.
        """
        self.parent.remove_block(self)



"""
Module for ELBv2 Target Group resources.
"""
from .base_control import Describer, ResourceLister
from .common import Common


class TargetGroupDescriber(Describer):
    """
    Describer control for ELBv2 Target Groups.
    """

    prefix = "tg_browser"
    title = "Target Group"

    resource_type = "target group"
    main_provider = "elbv2"
    category = "ELB v2"
    subcategory = "Target Group"
    describe_method = "describe_target_groups"
    describe_kwarg_name = "TargetGroupArns"
    describe_kwarg_is_list = True
    object_path = ".TargetGroups[0]"
    default_entry_key = "arn"

    def populate_entry(self, **kwargs):
        super().populate_entry(**kwargs)
        try:
            self.name = kwargs["entry"]["name"]
        except KeyError:
            self.name = self.entry_id

    def title_info(self):
        return self.name


class TargetGroupResourceLister(ResourceLister):
    """
    Lister control for ELBv2 Target Groups.
    """

    prefix = "target_group_list"
    title = "Target Groups"
    command_palette = ["tg", "targetgroup"]

    resource_type = "target group"
    main_provider = "elbv2"
    category = "ELB v2"
    subcategory = "Target Group"
    list_method = "describe_target_groups"
    item_path = ".TargetGroups"
    columns = {
        "name": {
            "path": ".TargetGroupName",
            "size": 30,
            "weight": 0,
            "sort_weight": 0,
        },
        "protocol": {"path": ".Protocol", "size": 10, "weight": 1},
        "port": {"path": ".Port", "size": 10, "weight": 2},
        "target type": {"path": ".TargetType", "size": 10, "weight": 3},
        "arn": {"path": ".TargetGroupArn", "hidden": True},
    }
    describe_command = TargetGroupDescriber.opener
    primary_key = "arn"

    def title_info(self):
        return self.rule["arn"] if self.rule is not None else None

    def __init__(self, *args, rule_entry=None, **kwargs):
        self.rule = rule_entry
        if rule_entry is not None:
            arns = []
            raw = self.rule.controller_data
            try:
                arns = [
                    i["TargetGroupArn"]
                    for i in raw["Actions"][-1]["ForwardConfig"]["TargetGroups"]
                ]
                self.list_kwargs = {"TargetGroupArns": arns}
            except KeyError:
                self.rule = None
                Common.Session.set_message(
                    "Rule is not configured for forwarding.",
                    Common.color("message_error"),
                )
        super().__init__(*args, **kwargs)



"""
Mdoule for controls related to load balanching.
"""
from .base_control import Describer, ResourceLister, ResourceRefByClass


class LBDescriber(Describer):
    """
    Describer control for v2 load balancers.
    """

    prefix = "lb_browser"
    title = "Load Balancer"

    resource_type = "load balancer"
    main_provider = "elbv2"
    category = "ELB v2"
    subcategory = "Load Balancer"
    describe_method = "describe_load_balancers"
    describe_kwarg_name = "Names"
    describe_kwarg_is_list = True
    object_path = ".LoadBalancers[0]"


def _elbv2_determine_scheme(result, *args):
    """
    Column callback for determining load balancer scheme.
    """
    if result["Scheme"] == "internet-facing":
        return "public"
    return "private"


class LBResourceLister(ResourceLister):
    """
    Lister control for v2 load balancers.
    """

    prefix = "lb_list"
    title = "Load Balancers"
    command_palette = ["lb", "elbv2", "loadbalancing"]

    resource_type = "load balancer"
    main_provider = "elbv2"
    category = "ELB v2"
    subcategory = "Load Balancer"
    list_method = "describe_load_balancers"
    item_path = ".LoadBalancers"
    columns = {
        "name": {
            "path": ".LoadBalancerName",
            "size": 30,
            "weight": 0,
            "sort_weight": 0,
        },
        "type": {"path": ".Type", "size": 15, "weight": 1},
        "scheme": {
            "path": _elbv2_determine_scheme,
            "size": 8,
            "weight": 2,
        },
        "hostname": {
            "path": ".DNSName",
            "size": 120,
            "weight": 3,
        },
        "arn": {"path": ".LoadBalancerArn", "hidden": True},
    }
    describe_command = LBDescriber.opener
    open_command = ResourceRefByClass("ListenerResourceLister")
    open_selection_arg = "lb"



"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as readme:
    ld = readme.read()

setup(
    name="awsc",
    url="https://github.com/baliame/awsc",
    version="0.4.0rc10",
    description="AWS Commander",
    long_description=ld,
    long_description_content_type="text/markdown",
    author="baliame",
    author_email="akos.toth@cheppers.com",  # Optional
    classifiers=[  # Optional
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
    ],
    package_dir={"": "src"},  # Optional
    packages=find_packages(where="src"),  # Required
    python_requires=">=3.8, <4",
    install_requires=[
        "blessed==1.17.12",
        "cryptography>=39.0.2",
        "python-magic==0.4.27",
        "chardet==5.0.0",
        "pyyaml==5.3.1",
        "boto3==1.17.59",
        "jq==1.4.0",
        "pyperclip==1.8.1",
        "pygments>=2.14.0",
        "numpy",
        "packaging",
        "watchdog",
    ],
    entry_points={  # Optional
        "console_scripts": [
            "awsc=awsc:main",
            "awsc-creds=awsc:cred_helper",
        ],
    },
)



"""
Module for Elastic Beanstalk-related resources.
"""

from .base_control import (
    AttributeResolver,
    Describer,
    DescriberEntryAttribute,
    DescriberKwarg,
    FieldValue,
    ForceFlag,
    FString,
    IndexResolver,
    ResourceLister,
    ResourceRefByClass,
    ResourceRefByCommand,
    SelectionAttribute,
    SingleRelationLister,
    SingleSelectorDialog,
    TemplateDict,
)
from .common import Common
from .termui.dialog import DialogFieldCheckbox
from .termui.ui import ControlCodes


class EBApplicationDescriber(Describer):
    """
    Describer control for Beanstalk Application resources.
    """

    prefix = "eb_application_browser"
    title = "Beanstalk Application"

    resource_type = "application"
    main_provider = "elasticbeanstalk"
    category = "Elastic Beanstalk"
    subcategory = "Application"
    describe_method = "describe_applications"
    describe_kwarg_name = "ApplicationNames"
    describe_kwarg_is_list = True
    object_path = ".Applications[0]"


class EBApplicationResourceLister(ResourceLister):
    """
    Lister control for Beanstalk Application resources.
    """

    prefix = "eb_application_list"
    title = "Beanstalk Applications"
    command_palette = ["ebapplication", "ebapp", "elasticbeanstalkapplication"]

    resource_type = "application"
    main_provider = "elasticbeanstalk"
    category = "Elastic Beanstalk"
    subcategory = "Application"
    list_method = "describe_applications"
    item_path = ".Applications"
    columns = {
        "name": {"path": ".ApplicationName", "size": 20, "weight": 0, "sort_weight": 0},
        "description": {"path": ".Description", "size": 50, "weight": 1},
        "arn": {"path": ".ApplicationArn", "hidden": True},
    }
    describe_command = EBApplicationDescriber.opener
    open_command = ResourceRefByCommand("elasticbeanstalkenvironment")
    open_selection_arg = "app"

    @ResourceLister.Autohotkey(ControlCodes.D, "Delete application", True)
    def delete_application(self, *args, **kwargs):
        """
        Hotkey definition for deleting an application.
        """
        self.confirm_template(
            "delete_application",
            TemplateDict(
                {
                    "ApplicationName": SelectionAttribute("name"),
                    "TerminateEnvByForce": ForceFlag(),
                }
            ),
            can_force=True,
        )(self.selection)


class EBApplicationVersionDescriber(Describer):
    """
    Describer control for Beanstalk Application Version resources.
    """

    prefix = "eb_application_version_browser"
    title = "Beanstalk Application Version"

    resource_type = "appversion"
    main_provider = "elasticbeanstalk"
    category = "Elastic Beanstalk"
    subcategory = "Appversion"
    describe_method = "describe_application_versions"
    describe_kwarg_name = "VersionLabels"
    describe_kwarg_is_list = True
    describe_kwargs_override = {
        "ApplicationName": DescriberEntryAttribute("application")
    }
    object_path = ".Applications[0]"


@ResourceLister.Autocommand("EBApplicationLister", "v", "View Versions", "application")
class EBApplicationVersionResourceLister(ResourceLister):
    """
    Lister control for Beanstalk Application Version resources.
    """

    prefix = "eb_application_version_list"
    title = "Beanstalk Application Versions"
    command_palette = [
        "ebapplicationversion",
        "ebappversion",
        "elasticbeanstalkapplicationversion",
    ]

    resource_type = "appversion"
    main_provider = "elasticbeanstalk"
    category = "Elastic Beanstalk"
    subcategory = "Appversion"
    list_method = "describe_application_versions"
    item_path = ".ApplicationVersions"
    columns = {
        "application": {
            "path": ".ApplicationName",
            "size": 20,
            "weight": 0,
            "sort_weight": 0,
        },
        "version": {"path": ".VersionLabel", "size": 20, "weight": 1, "sort_weight": 1},
        "status": {"path": ".Status", "size": 20, "weight": 2},
        "description": {"path": ".Description", "size": 50, "weight": 3},
        "name": {"path": ".VersionLabel", "hidden": True},
    }
    describe_command = EBApplicationVersionDescriber.opener

    def __init__(self, *args, application=None, **kwargs):
        self.application = application
        if application is not None:
            self.list_kwargs = {"ApplicationName": self.application["name"]}
        super().__init__(*args, **kwargs)
        self.confirm_template(
            "delete_application_version",
            TemplateDict(
                {
                    "ApplicationName": SelectionAttribute("name"),
                    "VersionLabel": SelectionAttribute("version"),
                }
            ),
            can_force=True,
            hotkey=ControlCodes.D,
            hotkey_tooltip="Delete appversion",
        )


class EBEnvironmentDescriber(Describer):
    """
    Describer control for Beanstalk Environment resources.
    """

    prefix = "eb_environment_browser"
    title = "Beanstalk Environment"

    resource_type = "environment"
    main_provider = "elasticbeanstalk"
    category = "Elastic Beanstalk"
    subcategory = "Environment"
    describe_method = "describe_environments"
    describe_kwarg_name = "EnvironmentNames"
    describe_kwarg_is_list = True
    object_path = ".Environments[0]"
    default_entry_key = "arn"


@ResourceLister.Autocommand("EBEnvironmentLister", "h", "Environment health", "entry")
class EBEnvironmentHealthDescriber(Describer):
    """
    Describer control for Beanstalk Environment resource health state.
    """

    prefix = "eb_environment_health_browser"
    title = "Beanstalk Environment Health"

    resource_type = "environment"
    main_provider = "elasticbeanstalk"
    category = "Elastic Beanstalk"
    subcategory = "Environment Health"
    describe_method = "describe_environment_health"
    describe_kwarg_name = "EnvironmentName"
    describe_kwarg_is_list = True
    object_path = "."


class EBEnvironmentLister(ResourceLister):
    """
    Lister control for Beanstalk Environment resources.
    """

    prefix = "eb_environment_list"
    title = "Beanstalk Environments"
    command_palette = ["ebenvironment", "ebenv", "elasticbeanstalkenvironment"]

    resource_type = "environment"
    main_provider = "elasticbeanstalk"
    category = "Elastic Beanstalk"
    subcategory = "Environment"
    list_method = "describe_environments"
    item_path = ".Environments"
    columns = {
        "name": {"path": ".EnvironmentName", "size": 20, "weight": 0, "sort_weight": 1},
        "id": {"path": ".EnvironmentId", "size": 20, "weight": 1},
        "application": {
            "path": ".ApplicationName",
            "size": 20,
            "weight": 2,
            "sort_weight": 0,
        },
        "status": {"path": ".Status", "size": 15, "weight": 3, "sort_weight": 0},
        "health": {"path": ".Health", "size": 7, "weight": 4, "sort_weight": 0},
        "tier": {"path": ".Tier.Name", "size": 10, "weight": 5, "sort_weight": 0},
        "arn": {"path": ".EnvironmentArn", "hidden": True},
    }
    describe_command = EBEnvironmentDescriber.opener

    @ResourceLister.Autohotkey(ControlCodes.D, "Terminate environment", True, True)
    def delete_environment(self, _):
        """
        Hotkey callback for deleting an environment.
        """
        self.confirm_template(
            "terminate_environment",
            TemplateDict(
                {
                    "EnvironmentName": SelectionAttribute("name"),
                    "ForceTerminate": ForceFlag(),
                    "TerminateResources": FieldValue("terminate_resources_field"),
                }
            ),
            can_force=True,
            extra_fields={
                "terminate_resources_field": DialogFieldCheckbox(
                    "Terminate resources",
                    checked=False,
                    color=Common.color("modal_dialog_textfield_label"),
                    selected_color=Common.color("modal_dialog_textfield_selected"),
                )
            },
        )(self.selection)

    @ResourceLister.Autohotkey(ControlCodes.E, "Delete envconfig", True, True)
    def delete_environment_config(self, _):
        """
        Hotkey callback for deleting an environment configuration.
        """
        self.confirm_template(
            "delete_environment_config",
            TemplateDict(
                {
                    "EnvironmentName": SelectionAttribute("name"),
                    "ApplicationName": SelectionAttribute("application"),
                }
            ),
            resource_type="environment configuration",
        )(self.selection)

    @ResourceLister.Autohotkey(ControlCodes.B, "Rebuild environment", True, True)
    def rebuild_environment(self, _):
        """
        Hotkey callback for rebuilding an environment configuration.
        """
        self.confirm_template(
            "rebuild_environment",
            TemplateDict(
                {
                    "EnvironmentName": SelectionAttribute("name"),
                }
            ),
            action_name="Rebuild",
        )(self.selection)

    @ResourceLister.Autohotkey(ControlCodes.S, "Swap CNames", True, True)
    def swap_cnames(self, _):
        """
        Hotkey callback for swapping environment cnames.
        """
        SingleSelectorDialog.opener(
            f"Swap CNames of environment '{self.selection['name']}' with other environment",
            "environment",
            "swap",
            EBEnvironmentLister,
            self.do_swap,
            caller=self,
        )

    def do_swap(self, other_env):
        """
        Action callback for environment cname swap dialog.
        """
        if other_env == self.selection["name"]:
            Common.Session.error(
                "Source and destination environments are the same.",
                "Swap Cnames",
                "Elastic Beanstalk",
                resource=self.selection["name"],
            )
            return
        api_kwargs = {
            "SourceEnvironmentName": self.selection["name"],
            "DestinationEnvironmentName": other_env,
        }
        Common.generic_api_call(
            "elasticbeanstalk",
            "swap_environment_cnames",
            api_kwargs,
            "Swap Cnames",
            "Elastic Beanstalk",
            subcategory="Environment",
            success_template="Swapping cnames for environments {0}",
            resource=f"{self.selection['name']}, {other_env}",
        )
        self.refresh_data()

    def __init__(self, *args, app=None, **kwargs):
        self.app = app["name"]
        if app is not None:
            self.list_kwargs["ApplicationName"] = self.app
        super().__init__(*args, **kwargs)

    def title_info(self):
        return self.app


@ResourceLister.Autocommand("EBEnvironmentLister", "v", "View resources", "selection")
class EBEnvironmentResourceLister(SingleRelationLister):
    """
    Lister control for Beanstalk Environment related resources.
    """

    prefix = "eb_environment_resources"
    title = "Beanstalk Environment resources"

    resource_type = "beanstalk environment"
    main_provider = "elasticbeanstalk"
    category = "Elastic Beanstalk"
    subcategory = "Environment"
    describe_method = "describe_environment_resources"
    describe_kwargs = TemplateDict({"EnvironmentName": SelectionAttribute("name")})
    object_path = ".EnvironmentResources"
    resource_descriptors = [
        {
            "base_path": "[.AutoscalingGroups[].Name]",
            "type": "Auto Scaling Group",
            "describer": ResourceRefByClass("ASGDescriber"),
        },
        {
            "base_path": "[.Instances[].Id]",
            "type": "EC2 Instance",
            "describer": ResourceRefByClass("EC2Describer"),
        },
        {
            "base_path": "[.LaunchConfigurations[].Name]",
            "type": "Launch Configuration",
            "describer": ResourceRefByClass("LCDescriber"),
        },
        {
            "base_path": "[.LaunchTemplates[].Name]",
            "type": "Launch Templates",
        },
        {
            "base_path": "[.LoadBalancers[].Name]",
            "type": "Load Balancer",
            "describer": ResourceRefByClass("LBDescriber"),
        },
        {
            "base_path": "[.Triggers[].Name]",
            "type": "Trigger",
        },
        {
            "base_path": "[.Queues[].URL]",
            "type": "SQS Queue",
            "describer": ResourceRefByClass("SQSDescriber"),
        },
    ]

    def title_info(self):
        return f"Environment: {self.parent_selection['name']}"


class EBPlatformBranchLister(ResourceLister):
    """
    Lister control for Beanstalk Platform Branch resources.
    """

    prefix = "eb_platform_version_list"
    title = "Beanstalk Platform Branches"
    command_palette = [
        "platformbranch",
        "ebplatformbranch",
        "elasticbeanstalkplatformbranch",
    ]

    resource_type = "platform branch"
    main_provider = "elasticbeanstalk"
    category = "Elastic Beanstalk"
    subcategory = "Platform Branch"
    list_method = "list_platform_branches"
    item_path = ".PlatformBranchSummaryList"
    columns = {
        "platform": {
            "path": ".PlatformName",
            "size": 20,
            "weight": 0,
            "sort_weight": 0,
        },
        "branch": {"path": ".BranchName", "size": 20, "weight": 1, "sort_weight": 1},
        "lifecycle": {
            "path": ".LifecycleState",
            "size": 20,
            "weight": 2,
        },
        "order": {"path": ".BranchOrder", "size": 10, "weight": 3},
        "name": {"path": ".BranchName", "hidden": True},
    }
    describe_command = ResourceRefByCommand("elasticbeanstalkplatformversion")
    describe_selection_arg = "branch"


class EBPlatformVersionDescriber(Describer):
    """
    Describer control for Beanstalk Platform Versions.
    """

    prefix = "eb_platform_version_browser"
    title = "Beanstalk Platform Version"

    resource_type = "platform version"
    main_provider = "elasticbeanstalk"
    category = "Elastic Beanstalk"
    subcategory = "Platform Version"
    describe_method = "describe_platform_version"
    describe_kwarg_name = "PlatformArn"
    object_path = ".PlatformDescription"


class EBPlatformVersionLister(ResourceLister):
    """
    Lister control for Beanstalk Platform Versions.
    """

    prefix = "eb_platform_version_list"
    title = "Beanstalk Platform Versions"
    command_palette = [
        "platformversion",
        "ebplatformversion",
        "elasticbeanstalkplatformversion",
    ]

    resource_type = "platform version"
    main_provider = "elasticbeanstalk"
    category = "Elastic Beanstalk"
    subcategory = "Platform Version"
    list_method = "list_platform_versions"
    item_path = ".PlatformSummaryList"
    columns = {
        "owner": {
            "path": ".PlatformOwner",
            "size": 10,
            "weight": 0,
            "sort_weight": 2,
        },
        "os": {
            "path": ".OperatingSystemName",
            "size": 20,
            "weight": 1,
            "sort_weight": 0,
        },
        "os version": {
            "path": ".OperatingSystemVersion",
            "size": 10,
            "weight": 2,
            "sort_weight": 1,
        },
        "status": {"path": ".PlatformStatus", "size": 10, "weight": 3},
        "category": {"path": ".PlatformCategory", "size": 10, "weight": 4},
        "version": {"path": ".PlatformVersion", "size": 10, "weight": 5},
        "lifecycle": {"path": ".PlatformLifecycleState", "size": 10, "weight": 6},
        "name": {"path": ".PlatformArn", "hidden": True},
        "arn": {"path": ".PlatformArn", "hidden": True},
    }
    describe_command = EBPlatformVersionDescriber.opener

    def title_info(self):
        return f"Branch: {self.branch['name']}" if self.branch is not None else None

    def __init__(self, *args, branch=None, **kwargs):
        self.branch = branch
        if branch is not None:
            self.list_kwargs = {
                "Filters": [
                    {
                        "Type": "PlatformBranchName",
                        "Operator": "=",
                        "Values": [branch["name"]],
                    }
                ]
            }
        super().__init__(*args, **kwargs)


class EBInstanceHealthDescriber(Describer):
    """
    Describer control for instance health.
    """

    prefix = "eb_instance_health_browser"
    title = "Beanstalk Instance"

    resource_type = "instance health"
    main_provider = "elasticbeanstalk"
    category = "Elastic Beanstalk"
    subcategory = "Instance Health"
    describe_method = "describe_instances_health"
    describe_kwarg_name = "VersionLabels"
    describe_kwargs_override = {
        "EnvironmentName": DescriberKwarg(
            "caller",
            resolver=AttributeResolver("environment", resolver=IndexResolver("name")),
        )
    }
    object_path = FString(".InstanceList[] | select(.InstanceId=={entry_id}))")
    default_entry_key = "instance id"


@ResourceLister.Autocommand(
    "EBEnvironmentLister", "i", "View instance healths", "environment"
)
class EBInstanceHealthLister(ResourceLister):
    """
    Lister control for instance healths in Beanstalk Environments.
    """

    prefix = "eb_instance_health_list"
    title = "Beanstalk Instance Health"

    resource_type = "instance health"
    main_provider = "elasticbeanstalk"
    category = "Elastic Beanstalk"
    subcategory = "Instance Health"
    list_method = "describe_instances_health"
    item_path = ".InstanceHealthList"
    columns = {
        "instance id": {
            "path": ".InstanceId",
            "size": 20,
            "weight": 0,
            "sort_weight": 0,
        },
        "az": {"path": ".AvailabilityZone", "size": 20, "weight": 1},
        "type": {"path": ".InstanceType", "size": 20, "weight": 2},
        "status": {"path": ".HealthStatus", "size": 20, "weight": 3},
        "color": {"path": ".Color", "size": 20, "weight": 4},
        "name": {"path": ".InstanceId", "hidden": True},
    }
    describe_command = EBInstanceHealthDescriber.opener

    def __init__(self, *args, environment=None, **kwargs):
        self.environment = environment
        self.list_kwargs = {"EnvironmentName": environment["name"]}
        super().__init__(*args, **kwargs)



"""
Module for Launch Configuration resources.
"""
from .base_control import Describer, ResourceLister, ResourceRefByClass


class LCDescriber(Describer):
    """
    Describer control for autoscaling launch configurations.
    """

    prefix = "lc_browser"
    title = "Launch Configuration"

    resource_type = "launch configuration"
    main_provider = "autoscaling"
    category = "Autoscaling"
    subcategory = "Launch Configuration"
    describe_method = "describe_launch_configurations"
    describe_kwarg_name = "LaunchConfigurationNames"
    describe_kwarg_is_list = True
    object_path = ".LaunchConfigurations[0]"


class LCResourceLister(ResourceLister):
    """
    Lister control for autoscaling launch configurations.
    """

    prefix = "lc_list"
    title = "Launch Configurations"
    command_palette = ["lc", "launchconfiguration"]

    resource_type = "launch configuration"
    main_provider = "autoscaling"
    category = "Autoscaling"
    subcategory = "Launch Configuration"
    list_method = "describe_launch_configurations"
    item_path = ".LaunchConfigurations"
    columns = {
        "name": {
            "path": ".LaunchConfigurationName",
            "size": 30,
            "weight": 0,
            "sort_weight": 0,
        },
        "image id": {"path": ".ImageId", "size": 20, "weight": 1},
        "instance type": {
            "path": ".InstanceType",
            "size": 20,
            "weight": 2,
        },
    }
    describe_command = LCDescriber.opener
    open_command = ResourceRefByClass("ASGResourceLister")
    open_selection_arg = "lc"



"""
Module for IAM group resources.
"""

from .base_control import (
    Describer,
    ResourceLister,
    SelectionAttribute,
    SelectorTemplateField,
    SingleNameDialog,
    TemplateDict,
)
from .common import Common
from .resource_iam import InlinePolicyAttacher
from .termui.common import Commons
from .termui.ui import ControlCodes


class GroupDescriber(Describer):
    """
    Describer control for IAM groups.
    """

    prefix = "group_browser"
    title = "Group"

    resource_type = "group"
    main_provider = "iam"
    category = "IAM"
    subcategory = "Group"
    describe_method = "get_group"
    describe_kwarg_name = "GroupName"
    object_path = ".Group"


@ResourceLister.Autocommand("UserLister", "g", "List groups for user", "user")
class GroupLister(ResourceLister):
    """
    Lister control for IAM groups.
    """

    prefix = "group_list"
    title = "Groups"
    command_palette = ["group", "groups"]

    resource_type = "group"
    main_provider = "iam"
    category = "IAM"
    subcategory = "Group"
    item_path = ".Groups"
    columns = {
        "name": {
            "path": ".GroupName",
            "size": 25,
            "weight": 0,
            "sort_weight": 0,
        },
        "id": {"path": ".GroupId", "size": 15, "weight": 1},
        "path": {"path": ".Path", "size": 40, "weight": 2},
        "arn": {"path": ".Arn", "hidden": True},
    }
    describe_command = GroupDescriber.opener
    open_command = "m"

    def title_info(self):
        return f"User: {self.user['name']}" if self.user is not None else None

    @ResourceLister.Autohotkey("a", "Attach policy", True)
    def attach_policy(self, _):
        """
        Hotkey callback for attaching a policy to a group.
        """
        self.selector_template(
            "attach_group_policy",
            TemplateDict(
                {
                    "GroupName": SelectionAttribute("name"),
                    "PolicyArn": SelectorTemplateField(),
                }
            ),
            "Attach policy to group",
            "attach",
            "PolicyLister",
            resource_type="policy",
            success_template="Attaching policy {PolicyArn} to group {GroupName}",
        )(self.selection)

    def create_group(self, _):
        """
        Hotkey callback for creating a new group.
        """
        SingleNameDialog(
            self.parent,
            "Create new group",
            self.do_create,
            label="Name:",
            what="name",
            subject="group",
            default="",
            caller=self,
            accepted_inputs=Commons.alphanum_with_symbols("+=,.@_-"),
        )

    def do_create(self, name):
        """
        Action callback for creating a new group.
        """
        Common.generic_api_call(
            "iam",
            "create_group",
            {"GroupName": name},
            "Create Group",
            "IAM",
            subcategory="Group",
            success_template="Creating group {GroupName}",
            resource=self.selection["name"],
        )
        self.refresh_data()

    def do_remove(self, **kwargs):
        """
        Action callback for removing a user from a group.
        """
        Common.generic_api_call(
            "iam",
            "remove_user_from_group",
            {"GroupName": self.selection["name"], "UserName": self.user["name"]},
            "Remove User From Group",
            "IAM",
            subcategory="Group",
            success_template="Removing user {UserName} from group {GroupName}",
            resource=self.user["name"],
        )
        self.refresh_data()

    def add_user_to_group(self, _):
        """
        Hotkey callback for adding a user to a group.
        """
        self.selector_template(
            "add_user_to_group",
            TemplateDict(
                {
                    "GroupName": SelectionAttribute("name"),
                    "UserName": SelectorTemplateField(),
                }
            ),
            "Add user to group",
            "add",
            "UserLister",
            resource_type="user",
            success_template="Adding user {UserName} to group {GroupName}",
        )(self.selection)

    def __init__(self, *args, user=None, **kwargs):
        self.user = user
        self.list_method = "list_groups_for_user" if user is not None else "list_groups"
        self.item_path = ".Groups"
        self.list_kwargs = {"UserName": self.user["name"]} if user is not None else {}

        super().__init__(*args, **kwargs)
        self.inliner = InlinePolicyAttacher(self, "group")
        if self.user is None:
            self.confirm_template(
                "delete_group",
                TemplateDict(
                    {
                        "GroupName": SelectionAttribute("name"),
                    }
                ),
                hotkey=ControlCodes.D,
                hotkey_tooltip="Delete Group",
            )
            self.add_hotkey("j", self.add_user_to_group, "Add User", True)
            self.add_hotkey("n", self.create_group, "Create Group")
        else:
            self.confirm_template(
                "remove_user_from_group",
                TemplateDict(
                    {
                        "UserName": self.user["name"],
                        "GroupName": SelectionAttribute("name"),
                    }
                ),
                resource_type="user",
                resource_identifier=self.user["name"],
                from_what="group",
                from_what_name=SelectionAttribute("name"),
                action_name="Remove",
                hotkey=ControlCodes.D,
                hotkey_tooltip="Remove from Group",
            )



"""
Module for working with ARNs (Amazon Resource Names).
"""


class ARN:
    """
    Represents an Amazon Resource Name.

    Attributes
    ----------
    partition : str
        The partition of the ARN.
    service : str
        The AWS service identified by the ARN.
    region : str
        The region of the resource identified by ARN, if applicable.
    account_id : str
        The AWS account the resource belongs to, if applicable.
    resource_type : str
        The resource type of the resource, if present in the ARN.
    resource_id : str
        The ID of the resource identified by the ARN.
    """

    def __init__(self, string):
        """
        Initializes an ARN object from an existing ARN string.

        Parameters
        ----------
        string : str
            The ARN string to initialize this object from.
        """
        parts = string.split(":")
        if len(parts) < 6:
            raise ValueError(
                "ARN does not contain enough components for full identification"
            )
        self.partition = parts[1] if parts[1] != "" else "aws"
        if parts[2] == "":
            raise ValueError("ARN is missing service identifier.")
        self.service = parts[2]
        self.region = parts[3]
        self.account_id = parts[4]
        if len(parts) == 6:
            test = parts[5].split("/")
            if len(test) > 1:
                self.resource_type = test[0]
                self.resource_id = "/".join(test[1:])
            else:
                self.resource_id = parts[5]
                self.resource_type = ""
        else:
            self.resource_type = parts[5]
            self.resource_id = ":".join(parts[6:])

    @property
    def resource(self):
        """
        Returns the resource portion of the ARN.

        If the resource type is present in the ARN, the full resource section including the type will be returned.

        Returns
        -------
        str
            The full resource section of the ARN.
        """
        if self.resource_type != "":
            return f"{self.resource_type}/{self.resource_id}"
        return self.resource_id

    @property
    def resource_id_first(self):
        """
        Returns the first resource identifier in the ARN.

        In some cases, the ARN may contain multiple resource IDs separated by slashes. This property is useful
        if only the first ID is required.

        Returns
        -------
        str
            The resource ID until the first slash encountered in the string.
        """
        return self.resource_id.split("/")[0]

    def __str__(self):
        """
        Converts the ARN object into an ARN string.

        Returns
        -------
        str
            The string representation of the ARN.
        """
        return f"arn:{self.partition}:{self.service}:{self.region}:{self.account_id}:{self.resource}"



"""
Module for database subnet groups.
"""
from .base_control import Describer, ResourceLister


class DBSubnetGroupDescriber(Describer):
    """
    Describer control for database subnets.
    """

    prefix = "db_subnet_group_browser"
    title = "DB Subnet Group"

    resource_type = "db subnet group"
    main_provider = "rds"
    category = "RDS"
    subcategory = "DB Subnet Groups"
    describe_method = "describe_db_subnet_groups"
    describe_kwarg_name = "DBSubnetGroupName"
    describe_kwarg_is_list = True
    object_path = ".DBSubnetGroups[0]"


class DBSubnetGroupResourceLister(ResourceLister):
    """
    Lister control for database subnets.
    """

    prefix = "db_subnet_group_list"
    title = "DB Subnet Groups"
    command_palette = ["dsg", "dbsubnetgroup"]

    resource_type = "db subnet group"
    main_provider = "rds"
    category = "RDS"
    subcategory = "DB Subnet Groups"
    list_method = "describe_db_subnet_groups"
    item_path = ".DBSubnetGroups"
    columns = {
        "vpc": {"path": ".VpcId", "size": 30, "weight": 0, "sort_weight": 0},
        "name": {
            "path": ".DBSubnetGroupName",
            "size": 30,
            "weight": 1,
            "sort_weight": 1,
        },
        "status": {"path": ".SubnetGroupStatus", "size": 30, "weight": 2},
    }
    describe_command = DBSubnetGroupDescriber.opener
    open_command = "t"



"""
Module for VPC routing resources.
"""
import json

from .base_control import (
    Describer,
    GenericDescriber,
    ResourceLister,
    tagged_column_generator,
)
from .common import Common


def _route_determine_gateway_type(self, entry):
    """
    Column callback for determining the type of a gateway at the tail end of a route.
    """
    if "NatGatewayId" in entry:
        return "NAT"
    if "InstanceId" in entry:
        return "Instance"
    if "TransitGatewayId" in entry:
        return "Transit"
    if "LocalGatewayId" in entry:
        return "Local"
    if "CarrierGatewayId" in entry:
        return "Carrier"
    if "VpcPeeringConnectionId" in entry:
        return "VPC Peering"
    if "EgressOnlyInternetGatewayId" in entry:
        return "Egress-Only"
    if entry["GatewayId"] == "local":
        return "VPC-Local"
    return "Internet"


def _route_determine_gateway(self, entry):
    """
    Column callback for determining the gateway ID at the tail end of a route.
    """
    if "NatGatewayId" in entry:
        return entry["NatGatewayId"]
    if "InstanceId" in entry:
        return entry["InstanceId"]
    if "TransitGatewayId" in entry:
        return entry["TransitGatewayId"]
    if "LocalGatewayId" in entry:
        return entry["LocalGatewayId"]
    if "CarrierGatewayId" in entry:
        return entry["CarrierGatewayId"]
    if "VpcPeeringConnectionId" in entry:
        return entry["VpcPeeringConnectionId"]
    if "EgressOnlyInternetGatewayId" in entry:
        return entry["EgressOnlyInternetGatewayId"]
    return entry["GatewayId"]


class RouteResourceLister(ResourceLister):
    """
    Lister control for VPC route resources.
    """

    prefix = "route_list"
    title = "Routes"
    command_palette = ["route"]

    resource_type = "route"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "VPC Route"
    list_method = "describe_route_tables"
    item_path = "[.RouteTables[] as $rt | $rt.Routes[] as $r | $r | .RouteTableId = $rt.RouteTableId]"
    columns = {
        "route table": {
            "path": ".RouteTableId",
            "size": 30,
            "weight": 0,
            "sort_weight": 0,
        },
        "gateway type": {
            "path": _route_determine_gateway_type,
            "size": 20,
            "weight": 1,
            "sort_weight": 1,
        },
        "gateway": {
            "path": _route_determine_gateway,
            "size": 30,
            "weight": 2,
        },
        "destination": {
            "path": ".DestinationCidrBlock",
            "size": 30,
            "weight": 3,
        },
        "state": {
            "path": ".State",
            "size": 10,
            "weight": 4,
        },
        "name": {
            "path": lambda _: "",
            "hidden": True,
        },
    }
    primary_key = None

    def title_info(self):
        if self.route_table is not None:
            return self.route_table["id"]
        return None

    def __init__(self, *args, route_table=None, **kwargs):
        if route_table is not None:
            self.route_table = route_table
            self.list_kwargs = {"RouteTableIds": [self.route_table["id"]]}
        super().__init__(*args, **kwargs)

    @ResourceLister.Autohotkey("d", "Describe", True)
    @ResourceLister.Autohotkey("KEY_ENTER", "Describe", True)
    def generic_describe(self, entry):
        """
        Custom describer for routes. We're inventing our own description here, AWS provides none.
        """
        Common.Session.push_frame(
            GenericDescriber.opener(
                **{
                    "describing": f"Route in route table {self.selection['route table']}",
                    "content": json.dumps(
                        self.selection.controller_data, sort_keys=True, indent=2
                    ),
                    "pushed": True,
                }
            )
        )


class RouteTableDescriber(Describer):
    """
    Describer control for VPC route tables.
    """

    prefix = "route_table_browser"
    title = "Route Table"

    resource_type = "route table"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "Route Table"
    describe_method = "describe_route_tables"
    describe_kwarg_name = "RouteTableIds"
    describe_kwarg_is_list = True
    object_path = ".RouteTables[0]"
    default_entry_key = "id"


def _route_table_determine_subnet_association(self, result, *args):
    if "Associations" not in result or len(result["Associations"]) == 0:
        return "<none>"
    if len(result["Associations"]) > 1:
        return "<multiple>"
    if "SubnetId" in result["Associations"][0]:
        return result["Associations"][0]["SubnetId"]
    return "<VPC default>"


@ResourceLister.Autocommand("SubnetResourceLister", "t", "View Route Table", "subnet")
class RouteTableResourceLister(ResourceLister):
    """
    Lister control for VPC route tables.
    """

    prefix = "route_table_list"
    title = "Route Tables"
    command_palette = ["rt", "routetable"]

    resource_type = "route table"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "VPC Route Table"
    list_method = "describe_route_tables"
    item_path = ".RouteTables"
    columns = {
        "id": {
            "path": ".RouteTableId",
            "size": 30,
            "weight": 0,
            "sort_weight": 1,
        },
        **tagged_column_generator("name", "name", weight=1, sort_weight=0, size=30),
        "vpc": {
            "path": ".VpcId",
            "size": 30,
            "weight": 2,
        },
        "type": {
            "path": _route_table_determine_subnet_association,
            "size": 30,
            "weight": 3,
        },
    }
    primary_key = "id"
    describe_command = RouteTableDescriber.opener
    open_command = RouteResourceLister.opener
    open_selection_arg = "route_table"

    def title_info(self):
        return self.subnet["id"] if self.subnet is not None else None

    def __init__(self, *args, subnet=None, **kwargs):
        self.subnet = subnet
        if subnet is not None:
            self.list_kwargs = {
                "Filters": [
                    {
                        "Name": "association.subnet-id",
                        "Values": [
                            self.subnet["id"],
                        ],
                    }
                ]
            }
        super().__init__(*args, **kwargs)



"""
Module for route 53 resources.
"""
import datetime
import json

from .base_control import (
    Describer,
    ListResourceDocumentCreator,
    ListResourceFieldsEditor,
    ResourceLister,
    SelectionAttribute,
    SelectionControllerDataAttribute,
    TemplateDict,
)
from .common import Common
from .termui.ui import ControlCodes


class R53RecordDescriber(Describer):
    """
    Describer control for Route 53 record sets.
    """

    prefix = "r53_record_browser"
    title = "Route53 Record"

    resource_type = "recordset"
    main_provider = "route53"
    category = "Route 53"
    subcategory = "Recordsets"
    describe_method = "list_resource_record_sets"
    describe_kwarg_name = "HostedZoneId"
    object_path = ".ResourceRecordSets[0]"
    default_entry_key = "hosted_zone_id"

    def populate_entry(self, **kwargs):
        super().populate_entry(**kwargs)
        self.record_type = kwargs["entry"]["entry"]
        self.record_name = kwargs["entry"]["name"]

    def populate_describe_kwargs(self, *args, **kwargs):
        super().populate_describe_kwargs(*args, **kwargs)
        self.describe_kwargs["StartRecordType"] = self.record_type
        self.describe_kwargs["StartRecordName"] = self.record_name

    def title_info(self):
        return f"{self.record_type} {self.record_name}"


def _r53_record_is_alias(result):
    return "AliasTarget" in result and result["AliasTarget"]["DNSName"] != ""


def _r53_record_determine_records(result):
    if _r53_record_is_alias(result):
        return result["AliasTarget"]["DNSName"]
    return ",".join([record["Value"] for record in result["ResourceRecords"]])


def _r53_record_determine_name(result):
    """
    Column callback for the name of a route 53 record.
    """
    return result["Name"].replace("\\052", "*")


class R53RecordLister(ResourceLister):
    """
    Lister control for route 53 recordsets.
    """

    prefix = "r53_record_list"
    title = "Route53 Recordsets"

    resource_type = "recordset"
    main_provider = "route53"
    category = "Route 53"
    subcategory = "Recordsets"
    list_method = "list_resource_record_sets"
    item_path = ".ResourceRecordSets"
    columns = {
        "entry": {
            "path": ".Type",
            "size": 5,
            "weight": 0,
            "sort_weight": 0,
        },
        "name": {
            "path": _r53_record_determine_name,
            "size": 30,
            "weight": 1,
            "sort_weight": 1,
        },
        "records": {
            "path": _r53_record_determine_records,
            "size": 60,
            "weight": 2,
        },
        "ttl": {
            "path": ".TTL",
            "size": 5,
            "weight": 3,
        },
    }
    primary_key = "name"
    describe_command = R53RecordDescriber.opener

    def title_info(self):
        return self.r53_entry["name"]

    def __init__(self, *args, r53_entry, **kwargs):
        self.r53_entry = r53_entry
        self.list_kwargs = {"HostedZoneId": self.r53_entry["id"]}
        super().__init__(*args, **kwargs)

    @ResourceLister.Autohotkey(ControlCodes.D, "Delete record", True)
    def delete_record(self, _):
        """
        Hotkey callback for deleting a record.
        """
        self.confirm_template(
            "change_resource_record_sets",
            TemplateDict(
                {
                    "HostedZoneId": self.r53_entry["id"],
                    "ChangeBatch": {
                        "Changes": [
                            {
                                "Action": "DELETE",
                                "ResourceRecordSet": {
                                    "Name": SelectionAttribute("name"),
                                    "Type": SelectionAttribute("entry"),
                                    "ResourceRecords": SelectionControllerDataAttribute(
                                        "ResourceRecords"
                                    ),
                                },
                            }
                        ]
                    },
                }
            ),
        )(self.selection)

    @ResourceLister.Autohotkey("n", "Add new record")
    def create(self, _):
        """
        Hotkey callback for adding a new record.
        """
        R53RecordCreator(self.r53_entry["id"]).edit()
        self.refresh_data()

    # TODO: Use new editor class if possible
    @ResourceLister.Autohotkey("e", "Edit", True)
    def edit(self, _):
        """
        Hotkey callback for editing an existing record set.
        """
        if self.selection is not None:
            raw = self.selection.controller_data
            if not _r53_record_is_alias(raw):
                content = "\n".join(
                    record["Value"] for record in raw["ResourceRecords"]
                )
                newcontent = Common.Session.textedit(content).strip(" \n\t")
                if content == newcontent:
                    Common.Session.set_message(
                        "Input unchanged.", Common.color("message_info")
                    )
                    return
                newcontent = newcontent.split("\n")
                Common.generic_api_call(
                    "route53",
                    "change_resource_record_sets",
                    {
                        "HostedZoneId": self.selection["hosted_zone_id"],
                        "ChangeBatch": {
                            "Changes": [
                                {
                                    "Action": "UPSERT",
                                    "ResourceRecordSet": {
                                        "Name": self.selection["name"],
                                        "Type": self.selection["entry"],
                                        "ResourceRecords": [
                                            {"Value": line} for line in newcontent
                                        ],
                                        "TTL": int(self.selection["ttl"]),
                                    },
                                }
                            ]
                        },
                    },
                    "Edit DNS record",
                    "Route53",
                    subcategory="Hosted Zone",
                    success_template="Modified DNS entry for zone {0}",
                    resource=self.selection["hosted_zone_id"],
                )

                self.refresh_data()
            else:
                Common.Session.set_message(
                    "Cannot edit aliased records", Common.color("message_info")
                )


class R53Describer(Describer):
    """
    Describer control for route 53 hosted zones.
    """

    prefix = "r53_browser"
    title = "Route53 Hosted Zone"

    def __init__(self, *args, entry_key="id", **kwargs):
        self.resource_key = "route53"
        self.describe_method = "get_hosted_zone"
        self.describe_kwarg_name = "Id"
        self.object_path = "."
        super().__init__(*args, entry_key=entry_key, **kwargs)


class R53ResourceLister(ResourceLister):
    """
    Lister control for route 53 hosted zones.
    """

    prefix = "r53_list"
    title = "Route53 Hosted Zones"
    command_palette = ["r53", "route53"]

    resource_type = "hosted zone"
    main_provider = "route53"
    category = "Route 53"
    subcategory = "Hosted Zones"
    list_method = "list_hosted_zones"
    item_path = ".HostedZones"
    columns = {
        "id": {
            "path": ".Id",
            "size": 30,
            "weight": 0,
            "sort_weight": 1,
        },
        "name": {
            "path": ".Name",
            "size": 30,
            "weight": 1,
            "sort_weight": 0,
        },
        "records": {
            "path": ".ResourceRecordSetCount",
            "size": 7,
            "weight": 2,
        },
    }
    primary_key = "id"
    describe_command = R53Describer.opener
    open_command = R53RecordLister.opener
    open_selection_arg = "r53_entry"

    @ResourceLister.Autohotkey(ControlCodes.D, "Delete hosted zone", True)
    def delete_hosted_zone(self, _):
        """
        Hotkey callback for deleting a hosted zone.
        """
        self.confirm_template(
            "delete_hosted_Zone",
            TemplateDict(
                {
                    "Id": SelectionAttribute("id"),
                }
            ),
        )(self.selection)

    @ResourceLister.Autohotkey("n", "Create hosted zone")
    def create(self, _):
        """
        Hotkey callback for creating a new hosted zone.
        """
        ListResourceDocumentCreator(
            "route53",
            "created_hosted_zone",
            None,
            initial_document={
                "Name": "",
                "VPC": {
                    "VPCRegion": "",
                    "VPCId": "",
                },
                "CallerReference": datetime.datetime.now().isoformat(),
                "HostedZoneConfig": {
                    "Comment": "",
                    "PrivateZone": False,
                },
                "DelegationSetId": "",
            },
        ).edit()


class R53RecordCreator(ListResourceDocumentCreator):
    """
    Custom creator class for creating Route 53 record sets. Route 53 has a strongly non-standard API.
    """

    def __init__(self, hosted_zone_id, *args, **kwargs):
        self.hosted_zone_id = hosted_zone_id
        super().__init__(
            "route53",
            "change_resource_record_sets",
            None,
            initial_document={
                "Name": "",
                "Type": "A",
                "ResourceRecords": [
                    {
                        "Value": "",
                    }
                ],
                "TTL": 30,
            },
        )

    def generate_kwargs(self, selection, newcontent):
        update_data = json.loads(newcontent)
        return {
            "HostedZoneId": self.hosted_zone_id,
            "ChangeBatch": {
                "Changes": [{"Action": "CREATE", "ResourceRecordSet": update_data}],
            },
        }


def _r53_healthcheck_determine_uri(result):
    """
    Column callback for fetching the URI for a route 53 healtcheck.
    """
    hcc = result["HealthCheckConfig"]
    if hcc["Type"] == "CLOUDWATCH_METRIC":
        cwac = result["CloudwatchAlarmConfiguration"]
        if cwac["ComparisonOperator"] == "GreaterThanOrEqualToThreshold":
            op = ">="
        elif cwac["ComparisonOperator"] == "GreaterThanThreshold":
            op = ">"
        elif cwac["ComparisonOperator"] == "LessThanOrEqualToThreshold":
            op = "<="
        elif cwac["ComparisonOperator"] == "LessThanThreshold":
            op = "<"
        else:
            op = "?"
        return f"<cloudwatch>: {cwac['MetricName']} {op} {cwac['Threshold']}"
    if hcc["Type"] == "CALCULATED":
        return "<calculated>"
    if hcc["Type"] == "RECOVERY_CONTROL":
        return "<recovery control>"
    if ["IPAddress"] in hcc and hcc["IPAddress"] != "":
        host = hcc["IPAddress"]
    else:
        host = hcc["FullyQualifiedDomainName"]
    if hcc["Type"] in ["HTTP", "HTTP_STR_MATCH"]:
        protocol = "http"
    elif hcc["Type"] in ["HTTPS", "HTTPS_STR_MATCH"]:
        protocol = "https"
    elif hcc["Type"] == "TCP":
        protocol = "tcp"
    else:
        protocol = f"unknown({hcc['Type']})"
    return f"{protocol}://{host}:{hcc['Port']}{hcc['ResourcePath']}"


def _r53_healthcheck_determine_status(result):
    """
    Column callback for determining a healthcheck status.
    """
    resp = Common.generic_api_call(
        "route53",
        "get_health_check_status",
        {"HealthCheckId": result["Id"]},
        "Get Health Check Status",
        "Route53",
        subcategory="Health Check",
        resource=result["Id"],
    )
    if resp["Success"]:
        return resp["Response"]["HealthCheckObservations"]["StatusReport"]["Status"]
    return "<n/a>"


class R53HealthCheckLister(ResourceLister):
    """
    Lister control for Route 53 health checks.
    """

    prefix = "r53_health_check_list"
    title = "Route53 Health Checks"
    # NOTE: healthcheck may conflict with other services
    # if it does, remove and only allow more explicit commands
    command_palette = ["r53healthcheck", "route53healthcheck", "healthcheck"]

    @ResourceLister.Autohotkey("n", "Create health check")
    def create_health_check(self, _):
        """
        Hotkey callback for creating a new health check.
        """
        ListResourceDocumentCreator(
            "route53",
            "create_health_check",
            "HealthCheckConfig",
            initial_document={
                "IPAddress": "string",
                "Port": 80,
                "Type": "HTTP",
                "ResourcePath": "/",
                "FullyQualifiedDomainName": "string",
                "SearchString": "string",
                "RequestInterval": 30,
                "FailureThreshold": 3,
                "MeasureLatency": False,
                "Inverted": False,
                "Disabled": False,
                "HealthThreshold": 0,
                "ChildHealthChecks": [
                    "string",
                ],
                "EnableSNI": True,
                "Regions": [
                    "us-east-1",
                ],
                "AlarmIdentifier": {"Region": "us-east-1", "Name": "string"},
                "InsufficientDataHealthStatus": "Unhealthy",
                "RoutingControlArn": "string",
            },
            as_json=True,
            static_fields={},
            message="Create successful",
        ).edit()

    @ResourceLister.Autohotkey("e", "Edit health check", True)
    def edit_health_check(self, _):
        """
        Hotkey callback for editing a health check.
        """
        ListResourceFieldsEditor(
            "route53",
            "get_health_check",
            ".HealthCheck.HealthCheckConfig",
            "update_health_check",
            "HealthCheckId",
            {
                "IPAddress": ".IPAddress",
                "Port": ".Port",
                "ResourcePath": ".ResourcePath",
                "FullyQualifiedDomainName": ".FullyQualifiedDomainName",
                "SearchString": ".SearchString",
                "FailureThreshold": ".FailureThreshold",
                "Inverted": ".Inverted",
                "Disabled": ".Disabled",
                "HealthThreshold": ".HealthThreshold",
                "ChildHealthChecks": ".ChildHealthChecks",
                "EnableSNI": ".EnableSNI",
                "Regions": ".Regions",
                "AlarmIdentifier": ".AlarmIdentifier",
                "InsufficientDataHealthStatus": ".InsufficientDataHealthStatus",
                "ResetElements": "[]",
            },
            entry_key="id",
            as_json=[
                "ChildHealthChecks",
                "Regions",
                "AlarmIdentifier",
                "ResetElements",
            ],
        ).edit(self.selection)

    @ResourceLister.Autohotkey(ControlCodes.D, "Delete health check", True)
    def delete_health_check(self, _):
        """
        Hotkey callback for deleting a health check.
        """
        self.confirm_template(
            "delete_health_check",
            TemplateDict(
                {
                    "HealthCheckId": SelectionAttribute("id"),
                }
            ),
        )(self.selection)

    resource_type = "route53 health check"
    main_provider = "route53"
    category = "Route 53"
    subcategory = "Health Check"
    list_method = "list_health_checks"
    item_path = ".HealthChecks"
    columns = {
        "id": {
            "path": ".Id",
            "size": 30,
            "weight": 0,
            "sort_weight": 1,
        },
        "uri": {
            "path": _r53_healthcheck_determine_uri,
            "size": 90,
            "weight": 1,
            "sort_weight": 1,
        },
        "inverted": {
            "path": ".HealthCheckConfig.Inverted",
            "size": 10,
            "weight": 2,
        },
        "disabled": {
            "path": ".HealthCheckConfig.Disabled",
            "size": 10,
            "weight": 3,
        },
        "status": {
            "path": _r53_healthcheck_determine_status,
            "size": 15,
            "weight": 4,
        },
        "name": {
            "path": ".Id",
            "size": 30,
            "hidden": True,
        },
    }
    primary_key = "id"
    # describe_command = R53HealthCheckDescriber.opener



"""
This module contains common termui features.
"""


class Commons:
    """
    TermUI Common class.

    This class holds the UI singleton as well as some default constants.

    Attributes
    ----------
    UIInstance : awsc.termui.ui.UI
        The UI singleton which has been instantiated for the terminal.
    TextfieldInputs : str
        All accepted inputs for a textfield which accepts any latin alphabet character, number, or most commonly used symbols.
    TextfieldInputsAlphaNum : str
        All accepted inputs for a textfield which accepts only latin alphabet characters and numbers.
    """

    UIInstance = None
    TextfieldInputs = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890/:@#.$=_-[]{}()<>|%!+~? "
    TextfieldInputsAlphaNum = (
        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890"
    )

    @classmethod
    def alphanum_with_symbols(cls, symbols):
        """
        Generates a set of accepted inputs for a textfield which accepts all latin alphabet characters and numbers but only a certain set of symbols.

        Returns
        -------
        str
            The set of accepted inputs for the textfield.
        """
        return f"{cls.TextfieldInputsAlphaNum}{symbols}"


def column_sizer(y0, y1, labels, data):
    """
    Determines column widths.

    Parameters
    ----------
    y0 : int
        Starting y.
    y1 : int
        Ending y.
    labels : list
        A list of labels.
    data : dict, optional
        A mapping of labels to values.

    Returns
    -------
    list(int)
        A list of widths for each column.
    """
    longest = []
    length = 0
    y = y0
    for name in labels:
        if data is not None and name not in data:
            continue
        display = name + ": "
        if len(display) > length:
            length = len(display)
        y += 1
        if y > y1:
            y = y0
            longest.append(length)
            length = 0
    if length > 0:
        longest.append(length)
    return longest



"""
Module for the color scheme configuration object.
"""
import yaml


class Scheme:
    """
    Color scheme holder class.

    Attributes
    ----------
    config : awsc.config.config.Configuration
        The parent configuration object instance.
    style_file : pathlib.Path
        The path to the color scheme yaml file.
    style : dict
        Holds the color scheme and border style. Refer to the contents of style_file for more information on structure.
    """

    def __init__(self, config):
        """
        Initializes a Scheme object.

        Parameters
        ----------
        config : awsc.config.config.Configuration
            The parent configuration object instance.
        """
        self.style_file = config.path / "style.yaml"
        self.config = config
        if not self.style_file.exists():
            self.create_default_config()
        else:
            self.parse_config()

    def backup_and_reset(self):
        """
        Restores the color scheme to default, creating a backup in the process.

        Sometimes, this is necessary as some keys may be added to the scheme during development.
        """
        print(
            "Restoring style scheme to defaults. Creating backup of pre-restore scheme."
        )
        backup = self.config.path / "style.yaml.bak"
        self.style_file.rename(backup)
        self.create_default_config()

    def create_default_config(self):
        """
        Creates the default first time style scheme, if style.yaml doesn't exist. This will overwrite style.yaml.
        """
        print("Creating first time style scheme...")
        self.style = {
            "colors": {
                "generic": {
                    "foreground": 220,
                    "background": 0,
                },
                "highlight": {
                    "foreground": 70,
                    "background": 0,
                },
                "highlight_selection": {
                    "foreground": 0,
                    "background": 70,
                },
                "info_display_title": {
                    "foreground": 70,
                    "background": 0,
                },
                "info_display_value": {
                    "foreground": 220,
                    "background": 0,
                },
                "hotkey_display_title": {
                    "foreground": 70,
                    "background": 0,
                },
                "hotkey_display_value": {
                    "foreground": 220,
                    "background": 0,
                },
                "modal_dialog_label_highlight": {
                    "foreground": 70,
                    "background": 0,
                },
                "modal_dialog_label": {
                    "foreground": 220,
                    "background": 0,
                },
                "generic_border": {
                    "foreground": 220,
                    "background": 0,
                },
                "modal_dialog_border": {
                    "foreground": 220,
                    "background": 0,
                },
                "border_title": {
                    "foreground": 111,
                    "background": 0,
                },
                "border_title_info": {
                    "foreground": 70,
                    "background": 0,
                },
                "modal_dialog_border_title": {
                    "foreground": 111,
                    "background": 0,
                },
                "modal_dialog_border_title_info": {
                    "foreground": 70,
                    "background": 0,
                },
                "column_title": {
                    "foreground": 0,
                    "background": 208,
                },
                "message_success": {
                    "foreground": 70,
                    "background": 0,
                },
                "message_info": {
                    "foreground": 33,
                    "background": 0,
                },
                "message_error": {
                    "foreground": 124,
                    "background": 0,
                },
                "error": {
                    "foreground": 124,
                    "background": 0,
                },
                "modal_dialog_error": {
                    "foreground": 124,
                    "background": 0,
                },
                "selection": {
                    "foreground": 0,
                    "background": 220,
                },
                "textfield_label": {
                    "foreground": 220,
                    "background": 0,
                },
                "textfield": {
                    "foreground": 0,
                    "background": 208,
                },
                "textfield_selection": {
                    "foreground": 0,
                    "background": 220,
                },
                "button": {
                    "foreground": 220,
                    "background": 0,
                },
                "button_selection": {
                    "foreground": 0,
                    "background": 220,
                },
                "context_list_generic": {
                    "foreground": 220,
                    "background": 0,
                },
                "context_list_selection": {
                    "foreground": 0,
                    "background": 220,
                },
                "context_list_border": {
                    "foreground": 220,
                    "background": 0,
                },
                "context_list_border_title": {
                    "foreground": 111,
                    "background": 0,
                },
                "context_list_heading": {
                    "foreground": 0,
                    "background": 208,
                },
                "search_bar_border": {
                    "foreground": 220,
                    "background": 0,
                },
                "search_bar_color": {
                    "foreground": 220,
                    "background": 0,
                },
                "search_bar_symbol_color": {
                    "foreground": 70,
                    "background": 0,
                },
                "search_bar_autocomplete_color": {
                    "foreground": 239,
                    "background": 0,
                },
                "search_bar_inactive_color": {
                    "foreground": 239,
                    "background": 0,
                },
                "command_bar_border": {
                    "foreground": 220,
                    "background": 0,
                },
                "command_bar_color": {
                    "foreground": 220,
                    "background": 0,
                },
                "command_bar_symbol_color": {
                    "foreground": 70,
                    "background": 0,
                },
                "command_bar_autocomplete_color": {
                    "foreground": 239,
                    "background": 0,
                },
                "command_bar_ok_color": {
                    "foreground": 33,
                    "background": 0,
                },
                "command_bar_error_color": {
                    "foreground": 124,
                    "background": 0,
                },
                "modal_dialog_textfield": {
                    "foreground": 0,
                    "background": 208,
                },
                "modal_dialog_textfield_selected": {
                    "foreground": 0,
                    "background": 220,
                },
                "modal_dialog_textfield_label": {
                    "foreground": 220,
                    "background": 0,
                },
                "syntax_highlight_token_text": {
                    "foreground": 220,
                    "background": 0,
                },
                "syntax_highlight_token_error": {
                    "foreground": 196,
                    "background": 0,
                },
                "syntax_highlight_token_other": {
                    "foreground": 243,
                    "background": 0,
                },
                "syntax_highlight_token_keyword": {
                    "foreground": 183,
                    "background": 0,
                },
                "syntax_highlight_token_keyword_constant": {
                    "foreground": 183,
                    "background": 0,
                },
                "syntax_highlight_token_keyword_declaration": {
                    "foreground": 183,
                    "background": 0,
                },
                "syntax_highlight_token_keyword_namespace": {
                    "foreground": 183,
                    "background": 0,
                },
                "syntax_highlight_token_keyword_pseudo": {
                    "foreground": 183,
                    "background": 0,
                },
                "syntax_highlight_token_keyword_reserved": {
                    "foreground": 183,
                    "background": 0,
                },
                "syntax_highlight_token_keyword_type": {
                    "foreground": 183,
                    "background": 0,
                },
                "syntax_highlight_token_literal": {
                    "foreground": 193,
                    "background": 0,
                },
                "syntax_highlight_token_literal_date": {
                    "foreground": 193,
                    "background": 0,
                },
                "syntax_highlight_token_literal_string": {
                    "foreground": 193,
                    "background": 0,
                },
                "syntax_highlight_token_literal_string_affix": {
                    "foreground": 193,
                    "background": 0,
                },
                "syntax_highlight_token_literal_string_backtick": {
                    "foreground": 193,
                    "background": 0,
                },
                "syntax_highlight_token_literal_string_char": {
                    "foreground": 193,
                    "background": 0,
                },
                "syntax_highlight_token_literal_string_delimiter": {
                    "foreground": 193,
                    "background": 0,
                },
                "syntax_highlight_token_literal_string_double": {
                    "foreground": 154,
                    "background": 0,
                },
                "syntax_highlight_token_literal_string_escape": {
                    "foreground": 193,
                    "background": 0,
                },
                "syntax_highlight_token_literal_string_heredoc": {
                    "foreground": 193,
                    "background": 0,
                },
                "syntax_highlight_token_literal_string_interpol": {
                    "foreground": 193,
                    "background": 0,
                },
                "syntax_highlight_token_literal_string_other": {
                    "foreground": 193,
                    "background": 0,
                },
                "syntax_highlight_token_literal_string_regex": {
                    "foreground": 193,
                    "background": 0,
                },
                "syntax_highlight_token_literal_string_single": {
                    "foreground": 193,
                    "background": 0,
                },
                "syntax_highlight_token_literal_string_symbol": {
                    "foreground": 193,
                    "background": 0,
                },
                "syntax_highlight_token_literal_number": {
                    "foreground": 111,
                    "background": 0,
                },
                "syntax_highlight_token_literal_number_bin": {
                    "foreground": 111,
                    "background": 0,
                },
                "syntax_highlight_token_literal_number_float": {
                    "foreground": 111,
                    "background": 0,
                },
                "syntax_highlight_token_literal_number_hex": {
                    "foreground": 111,
                    "background": 0,
                },
                "syntax_highlight_token_literal_number_integer": {
                    "foreground": 111,
                    "background": 0,
                },
                "syntax_highlight_token_literal_number_integer_long": {
                    "foreground": 111,
                    "background": 0,
                },
                "syntax_highlight_token_literal_number_oct": {
                    "foreground": 111,
                    "background": 0,
                },
                "syntax_highlight_token_operator": {
                    "foreground": 93,
                    "background": 0,
                },
                "syntax_highlight_token_operator_word": {
                    "foreground": 93,
                    "background": 0,
                },
                "syntax_highlight_token_punctuation": {
                    "foreground": 15,
                    "background": 0,
                },
                "syntax_highlight_token_punctuation_marker": {
                    "foreground": 15,
                    "background": 0,
                },
                "syntax_highlight_token_name": {
                    "foreground": 64,
                    "background": 0,
                },
                "syntax_highlight_token_name_tag": {
                    "foreground": 64,
                    "background": 0,
                },
                "syntax_highlight_token_name_attribute": {
                    "foreground": 64,
                    "background": 0,
                },
                "syntax_highlight_token_name_builtin": {
                    "foreground": 64,
                    "background": 0,
                },
                "syntax_highlight_token_name_builtin_pseudo": {
                    "foreground": 64,
                    "background": 0,
                },
                "syntax_highlight_token_name_class": {
                    "foreground": 64,
                    "background": 0,
                },
                "syntax_highlight_token_name_constant": {
                    "foreground": 64,
                    "background": 0,
                },
                "syntax_highlight_token_name_decorator": {
                    "foreground": 64,
                    "background": 0,
                },
                "syntax_highlight_token_name_entity": {
                    "foreground": 64,
                    "background": 0,
                },
                "syntax_highlight_token_name_exception": {
                    "foreground": 64,
                    "background": 0,
                },
                "syntax_highlight_token_name_function": {
                    "foreground": 64,
                    "background": 0,
                },
                "syntax_highlight_token_name_function_magic": {
                    "foreground": 64,
                    "background": 0,
                },
                "syntax_highlight_token_name_label": {
                    "foreground": 64,
                    "background": 0,
                },
                "syntax_highlight_token_name_namespace": {
                    "foreground": 64,
                    "background": 0,
                },
                "syntax_highlight_token_name_other": {
                    "foreground": 64,
                    "background": 0,
                },
                "syntax_highlight_token_name_variable": {
                    "foreground": 64,
                    "background": 0,
                },
                "syntax_highlight_token_name_variable_class": {
                    "foreground": 64,
                    "background": 0,
                },
                "syntax_highlight_token_name_variable_global": {
                    "foreground": 64,
                    "background": 0,
                },
                "syntax_highlight_token_name_variable_instance": {
                    "foreground": 64,
                    "background": 0,
                },
                "syntax_highlight_token_name_variable_magic": {
                    "foreground": 64,
                    "background": 0,
                },
                "dashboard_block_loading": {
                    "foreground": 33,
                    "background": 0,
                },
                "dashboard_block_error": {
                    "foreground": 124,
                    "background": 0,
                },
                "dashboard_block_label": {
                    "foreground": 220,
                    "background": 0,
                },
                "dashboard_block_information": {
                    "foreground": 75,
                    "background": 0,
                },
                "dashboard_block_positive": {
                    "foreground": 76,
                    "background": 0,
                },
                "dashboard_block_neutral": {
                    "foreground": 190,
                    "background": 0,
                },
                "dashboard_block_negative": {
                    "foreground": 209,
                    "background": 0,
                },
            },
            "borders": {
                "default": {
                    "horizontal": "─",
                    "vertical": "│",
                    "TL": "┌",
                    "TR": "┐",
                    "BL": "└",
                    "BR": "┘",
                },
                "search_bar": {
                    "horizontal": "─",
                    "vertical": "│",
                    "TL": "┌",
                    "TR": "┐",
                    "BL": "└",
                    "BR": "┘",
                },
                "modal": {
                    "horizontal": "─",
                    "vertical": "│",
                    "TL": "┌",
                    "TR": "┐",
                    "BL": "└",
                    "BR": "┘",
                },
                "resource_list": {
                    "horizontal": "─",
                    "vertical": "│",
                    "TL": "┌",
                    "TR": "┐",
                    "BL": "└",
                    "BR": "┘",
                },
            },
        }

        with self.style_file.open("w", encoding="utf-8") as file:
            file.write(yaml.dump(self.style))

    def __getitem__(self, item):
        """
        Pass-through to return the named item of the style dict.

        Parameters
        ----------
        item : str
            The key of style to return.

        Returns
        -------
        dict
            The value of style at item.
        """
        return self.style[item]

    def parse_config(self):
        """
        Parses the scheme configuration file. This will replace the contents of style.
        """
        with self.style_file.open("r", encoding="utf-8") as file:
            self.style = yaml.safe_load(file.read())
